@@ -13,11 +13,15 @@
 #ifndef EDITOR_H
 #define EDITOR_H
 #include "typesetter.hpp"
-#include "Widget/attribute_widget.hpp"
+#include "Widkit/simple_wk_widget.hpp"
 #include "server.hpp"
 #include "scheme.hpp"
 #include "drd_info.hpp"
-#define TEXMACS_COPYRIGHT (string("(c) 1999-2003 by Joris van der Hoeven"))
+#ifdef EXPERIMENTAL
+#include "../Style/Environment/environment.hpp"
+#include "../Style/Memorizer/memorizer.hpp"
+#endif
+#define TEXMACS_COPYRIGHT (string("(c) 1999-2006 by Joris van der Hoeven"))
 
 #define THE_CURSOR 1
 #define THE_FOCUS 2
@@ -26,13 +30,13 @@
 #define THE_SELECTION 16
 #define THE_DECORATIONS 32
 #define THE_EXTENTS 64
-#define THE_AUTOMATIC_SIZE 128
+#define THE_LOCUS 128
 
 class tm_buffer_rep;
 typedef tm_buffer_rep* tm_buffer;
 class server_rep;
 
-class editor_rep: public attribute_widget_rep {
+class editor_rep: public simple_widget_rep {
 public:
   server_rep* sv;   // the underlying texmacs server
 
@@ -43,6 +47,10 @@ protected:
   box         eb;   // box translation of tree
   path        rp;   // path to the root of the document in et
   path        tp;   // path of cursor in tree
+#ifdef EXPERIMENTAL
+  environment ste;  // environment for style rewriting
+  memorizer   mem;  // style converted document tree
+#endif
 
   /* exchanging information with the interface */
   virtual SI      get_window_height () = 0;
@@ -53,8 +61,6 @@ protected:
 
   /* exchanging information with the typesetter */
   virtual typesetter           get_typesetter () = 0;
-  virtual tree                 get_style () = 0;
-  virtual void                 set_style (tree t) = 0;
   virtual hashmap<string,tree> get_init () = 0;
   virtual hashmap<string,tree> get_fin () = 0;
   virtual void                 set_init (hashmap<string,tree> H= tree ("?"))=0;
@@ -88,8 +94,9 @@ protected:
   virtual void back_in_general (tree t, path p, bool forward) = 0;
 
   /* other protected subroutines */
-  virtual path tree_path (SI x, SI y, SI delta) = 0;
+  virtual path tree_path (path sp, SI x, SI y, SI delta) = 0;
   virtual void apply_changes () = 0;
+  virtual void animate () = 0;
   virtual void correct_concat (path p, int done=0) = 0;
   virtual path search_format (int& row, int& col) = 0;
   virtual void table_bound (path fp, int& i1, int& j1, int& i2, int& j2) = 0;
@@ -100,20 +107,22 @@ protected:
 
 public:
   editor_rep ();
-  editor_rep (server_rep* sv, display dis, tm_buffer buf);
+  editor_rep (server_rep* sv, tm_buffer buf);
   inline virtual ~editor_rep () {}
 
   /* public routines from edit_interface */
   virtual void suspend () = 0;
   virtual void resume () = 0;
   virtual int  get_pixel_size () = 0;
+  virtual void invalidate (SI x1, SI y1, SI x2, SI y2) = 0;
+  virtual void invalidate (rectangles rs) = 0;
   virtual void notify_change (int changed) = 0;
   virtual bool has_changed (int question) = 0;
-  virtual bool kbd_get_command (string cmd_s, string& help, command& cmd) = 0;
+  virtual int  idle_time (int event_type= ANY_EVENT) = 0;
+  virtual int  change_time () = 0;
   virtual void full_screen_mode (bool flag) = 0;
   virtual void before_menu_action () = 0;
   virtual void after_menu_action () = 0;
-  virtual void invalidate (SI x1, SI y1, SI x2, SI y2) = 0;
   virtual int  get_input_mode () = 0;
   virtual void set_input_mode (int mode) = 0;
   virtual void set_input_normal () = 0;
@@ -121,31 +130,38 @@ public:
   virtual bool in_search_mode () = 0;
   virtual bool in_replace_mode () = 0;
   virtual bool in_spell_mode () = 0;
+  virtual bool kbd_get_command (string cmd_s, string& help, command& cmd) = 0;
   virtual void key_press (string key) = 0;
   virtual void emulate_keyboard (string keys, string action= "") = 0;
-  virtual void show_keymaps () = 0;
   virtual bool complete_try () = 0;
   virtual void complete_start (string prefix, array<string> compls) = 0;
   virtual bool complete_keypress (string key) = 0;
-  virtual void mouse_any (string s, SI x, SI y, time_t t) = 0;
+  virtual void mouse_any (string s, SI x, SI y, int mods, time_t t) = 0;
   virtual void mouse_click (SI x, SI y) = 0;
   virtual bool mouse_extra_click (SI x, SI y) = 0;
   virtual void mouse_drag (SI x, SI y) = 0;
-  virtual void mouse_select (SI x, SI y) = 0;
+  virtual void mouse_select (SI x, SI y, int mods) = 0;
   virtual void mouse_paste (SI x, SI y) = 0;
   virtual void mouse_adjust (SI x, SI y) = 0;
   virtual void mouse_scroll (SI x, SI y, bool up) = 0;
   virtual cursor get_cursor () = 0;
-  virtual void set_message (string l, string r= "") = 0;
-  virtual void interactive (scheme_tree args, object cmd) = 0;
+  virtual void set_pointer (string name) = 0;
+  virtual void set_pointer (string curs_name, string mask_name) = 0;
+  virtual void set_message (string l, string r= "", bool temp= false) = 0;
+  virtual void recall_message () = 0;
 
   /* public routines from edit_cursor */
-  virtual path current_position () = 0;
-  virtual void go_to (SI x, SI y) = 0;
+  virtual path make_cursor_accessible (path p, bool forwards) = 0;
+  virtual void show_cursor_if_hidden () = 0;
+  virtual void go_to (SI x, SI y, bool absolute= true) = 0;
+  virtual void go_left_physical () = 0;
+  virtual void go_right_physical () = 0;
   virtual void go_left () = 0;
   virtual void go_right () = 0;
   virtual void go_up () = 0;
   virtual void go_down () = 0;
+  virtual void go_start_line () = 0;
+  virtual void go_end_line () = 0;
   virtual void go_page_up () = 0;
   virtual void go_page_down () = 0;
   virtual void go_to (path p) = 0;
@@ -156,28 +172,36 @@ public:
   virtual void go_to_here () = 0;
   virtual void go_start () = 0;
   virtual void go_end () = 0;
-  virtual void go_start_of (string what) = 0;
-  virtual void go_end_of (string what) = 0;
+  virtual void go_start_of (tree_label what) = 0;
+  virtual void go_end_of (tree_label what) = 0;
   virtual void go_start_with (string var, string val) = 0;
   virtual void go_end_with (string var, string val) = 0;
-  virtual void go_start_line () = 0;
-  virtual void go_end_line () = 0;
+  virtual void go_start_paragraph () = 0;
+  virtual void go_end_paragraph () = 0;
   virtual void go_to_label (string s) = 0;
   virtual tree get_labels () = 0;
 
   /* public routines from edit_graphics */
-  virtual bool   inside_graphics () = 0;
+  virtual bool   inside_graphics (bool b=true) = 0;
+  virtual bool   inside_active_graphics (bool b=true) = 0;
+  virtual bool   over_graphics (SI x, SI y) = 0;
   virtual tree   get_graphics () = 0;
-  virtual frame  find_frame () = 0;
+  virtual double get_x () = 0;
+  virtual double get_y () = 0;
+  virtual frame  find_frame (bool last= false) = 0;
+  virtual grid   find_grid () = 0;
   virtual void   find_limits (point& lim1, point& lim2) = 0;
+  virtual bool   find_graphical_region (SI& x1, SI& y1, SI& x2, SI& y2) = 0;
   virtual point  adjust (point p) = 0;
   virtual tree   find_point (point p) = 0;
   virtual tree   graphical_select (double x, double y) = 0;
+  virtual tree   graphical_select (double x1, double y1,
+				   double x2, double y2) = 0;
   virtual tree   get_graphical_object () = 0;
   virtual void   set_graphical_object (tree t) = 0;
   virtual void   invalidate_graphical_object () = 0;
-  virtual void   draw_graphical_object () = 0;
-  virtual bool   mouse_graphics (string s, SI x, SI y, time_t t) = 0;
+  virtual void   draw_graphical_object (renderer ren) = 0;
+  virtual bool   mouse_graphics (string s, SI x, SI y, int mods, time_t t) = 0;
 
   /* public routines from edit_typeset */
   virtual void     clear_local_info () = 0;
@@ -187,6 +211,9 @@ public:
   virtual bool     is_length (string s) = 0;
   virtual double   divide_lengths (string l1, string l2) = 0;
   virtual void     drd_update () = 0;
+#ifdef EXPERIMENTAL
+  virtual void     environment_update () = 0;
+#endif
   virtual bool     defined_at_cursor (string var_name) = 0;
   virtual bool     defined_at_init (string var_name) = 0;
   virtual bool     defined_in_init (string var_name) = 0;
@@ -200,16 +227,22 @@ public:
   virtual double   get_env_double (string var_name) = 0;
   virtual double   get_init_double (string var_name) = 0;
   virtual language get_env_language () = 0;
-  virtual tree     exec_texmacs (tree t, path p= 0) = 0;
+  virtual tree     exec_texmacs (tree t, path p) = 0;
+  virtual tree     exec_texmacs (tree t) = 0;
   virtual tree     exec_html (tree t, path p) = 0;
   virtual tree     exec_html (tree t) = 0;
+  virtual tree     exec_latex (tree t, path p) = 0;
+  virtual tree     exec_latex (tree t) = 0;
   virtual tree     texmacs_exec (tree t) = 0;
+  virtual tree     get_style () = 0;
+  virtual void     set_style (tree t) = 0;
   virtual void     init_style () = 0;
   virtual void     init_style (string style) = 0;
   virtual void     init_add_package (string package) = 0;
   virtual void     init_remove_package (string package) = 0;
   virtual void     init_env (string var, tree by) = 0;
   virtual void     init_default (string var) = 0;
+  virtual void     typeset_invalidate (path p) = 0;
   virtual void     typeset_invalidate_all () = 0;
 
   /* public routines from edit_modify */
@@ -218,30 +251,37 @@ public:
   virtual void remove (path p, int nr) = 0;
   virtual void split (path p) = 0;
   virtual void join (path p) = 0;
-  virtual void ins_unary (path p, tree_label op) = 0;
-  virtual void rem_unary (path p) = 0;
+  virtual void assign_node (path p, tree_label op) = 0;
+  virtual void insert_node (path p, tree t) = 0;
+  virtual void remove_node (path p) = 0;
   virtual void finished (path p) = 0;
   virtual void notify_assign (path p, tree u) = 0;
   virtual void notify_insert (path p, tree u) = 0;
   virtual void notify_remove (path p, int nr) = 0;
   virtual void notify_split (path p) = 0;
   virtual void notify_join (path p) = 0;
-  virtual void notify_ins_unary (path p, tree_label op) = 0;
-  virtual void notify_rem_unary (path p) = 0;
+  virtual void notify_assign_node (path p, tree_label op) = 0;
+  virtual void notify_insert_node (path p, tree t) = 0;
+  virtual void notify_remove_node (path p) = 0;
   virtual void post_notify (path p) = 0;
+  virtual void mark_undo_blocks () = 0;
+  virtual void remove_undo_mark () = 0;
+  virtual void add_undo_mark () = 0;
+  virtual void unredoable_undo () = 0;
+  virtual void forget_undo () = 0;
   virtual void undo () = 0;
   virtual void redo () = 0;
-  virtual void assign_diff (path p, tree u) = 0;
-  virtual int  position_new () = 0;
-  virtual void position_delete (int i) = 0;
-  virtual void position_set (int i, path p) = 0;
-  virtual path position_get (int i) = 0;
+  virtual observer position_new (path p) = 0;
+  virtual void position_delete (observer o) = 0;
+  virtual void position_set (observer o, path p) = 0;
+  virtual path position_get (observer o) = 0;
 
   /* public routines from edit_text */
   virtual void correct (path p) = 0;
   virtual bool insert_return () = 0;
   virtual void remove_return (path p) = 0;
   virtual void insert_tree (tree t, path p_in_t) = 0;
+  virtual void var_insert_tree (tree t, path p_in_t) = 0;
   virtual void insert_tree (tree t) = 0;
   virtual void remove_text (bool forward) = 0;
   virtual void remove_structure (bool forward) = 0;
@@ -259,8 +299,6 @@ public:
   virtual void make_htab (string spc) = 0;
   virtual void make_move (string x, string y) = 0;
   virtual void make_resize (string x1, string y1, string x2, string y2) = 0;
-  virtual void make_insertion (string s) = 0;
-  virtual void position_insertion (string what, bool flag) = 0;
   virtual void make_postscript (string file_name, bool link,
 				string w, string h,
 				string x1, string y1,
@@ -281,10 +319,6 @@ public:
   virtual void make_neg () = 0;
   virtual void make_tree () = 0;
 
-  virtual bool inside_tree () = 0;
-  virtual void branch_insert (bool at_right) = 0;
-  virtual void branch_delete (bool forward) = 0;
-
   /* public routines from edit_table */
   virtual void   make_table (int nr_rows=1, int nr_cols=1) = 0;
   virtual void   make_subtable (int nr_rows=1, int nr_cols=1) = 0;
@@ -292,8 +326,8 @@ public:
   virtual void   table_extract_format () = 0;
   virtual void   table_insert_row (bool forward) = 0;
   virtual void   table_insert_column (bool forward) = 0;
-  virtual void   table_delete_row (bool forward) = 0;
-  virtual void   table_delete_column (bool forward) = 0;
+  virtual void   table_remove_row (bool forward, bool flag= false) = 0;
+  virtual void   table_remove_column (bool forward, bool flag= false) = 0;
   virtual int    table_nr_rows () = 0;
   virtual int    table_nr_columns () = 0;
   virtual int    table_which_row () = 0;
@@ -306,13 +340,12 @@ public:
   virtual void   table_format_center () = 0;
   virtual void   table_row_decoration (bool forward) = 0;
   virtual void   table_column_decoration (bool forward) = 0;
+  virtual void   table_correct_block_content () = 0;
   virtual void   set_cell_mode (string mode) = 0;
   virtual string get_cell_mode () = 0;
   virtual void   cell_set_format (string var, string val) = 0;
   virtual string cell_get_format (string var) = 0;
   virtual void   cell_del_format (string var= "") = 0;
-  virtual void   cell_multi_paragraph (bool flag) = 0;
-  virtual bool   cell_is_multi_paragraph () = 0;
   virtual void   table_test () = 0;
 
   /* public routines from edit_dynamic */
@@ -324,6 +357,7 @@ public:
   virtual void insert_argument (path p, bool forward) = 0;
   virtual void insert_argument (bool forward) = 0;
   virtual void remove_argument (path p, bool forward) = 0;
+  virtual void remove_argument (bool forward) = 0;
   virtual void make_with (string var, string val) = 0;
   virtual void insert_with (path p, string var, tree val) = 0;
   virtual void remove_with (path p, string var) = 0;
@@ -359,8 +393,6 @@ public:
 
   virtual void session_go_up () = 0;
   virtual void session_go_down () = 0;
-  virtual void session_go_left () = 0;
-  virtual void session_go_right () = 0;
   virtual void session_go_page_up () = 0;
   virtual void session_go_page_down () = 0;
   virtual void session_remove (bool forward) = 0;
@@ -406,6 +438,7 @@ public:
   virtual void selection_get (path& start, path& end) = 0;
   virtual path selection_get_start () = 0;
   virtual path selection_get_end () = 0;
+  virtual path selection_get_path () = 0;
   virtual void selection_set (string key, tree t, bool persistant= false) = 0;
   virtual void selection_set (tree t) = 0;
   virtual void selection_set_start (path p= path()) = 0;
@@ -464,6 +497,7 @@ public:
   virtual void print_buffer (string first="1", string last="1000000") = 0;
   virtual void export_ps (url ps_name,
 			  string first="1", string last="1000000") = 0;
+  virtual array<int> print_snippet (url u, tree t) = 0;
   virtual void footer_eval (string s) = 0;
   virtual tree the_line () = 0;
   virtual tree the_root () = 0;
@@ -487,7 +521,7 @@ public:
 };
 
 class editor {
-  EXTEND_NULL(widget,editor);
+EXTEND_NULL(widget,editor);
 public:
   inline bool operator == (editor w) { return rep == w.rep; }
   inline bool operator != (editor w) { return rep != w.rep; }
