@@ -18,14 +18,20 @@
 #include "convert.hpp"
 #include "file.hpp"
 #include "sys_utils.hpp"
+#include "tmfs.hpp"
+#include "client_server.hpp"
 #include "analyze.hpp"
-#include "tm_layout.hpp"
+#include "tree_traverse.hpp"
+#include "tm_frame.hpp"
 #include "Concat/concater.hpp"
 #include "converter.hpp"
 #include "timer.hpp"
 #include "Metafont/tex_files.hpp"
 #include "Freetype/free_type.hpp"
 #include "Freetype/tt_file.hpp"
+#include "Bibtex/bibtex.hpp"
+#include "link.hpp"
+#include "dictionary.hpp"
 #include <string.h>
 #include <libguile.h>
 
@@ -49,10 +55,14 @@ scheme_dialect () {
 #else
 #ifdef GUILE_B
   return "guile-b";
+#else
+#ifdef GUILE_C
+  return "guile-c";
 #else
   return "unknown";
 #endif
 #endif
+#endif
 }
 
 string
@@ -80,6 +90,16 @@ win32_display (string s) {
   cout.flush ();
 }
 
+#ifdef GUILE_C
+#define SET_SMOB(smob,data,type)   \
+  SCM_NEWSMOB (smob, SCM_UNPACK (type), data);
+#else
+#define SET_SMOB(smob,data,type)   \
+  SCM_NEWCELL (smob);              \
+  SCM_SETCAR (smob, (SCM) (type)); \
+  SCM_SETCDR (smob, (SCM) (data));
+#endif
+
 /******************************************************************************
 * Direct access to scheme objects
 ******************************************************************************/
@@ -108,27 +128,31 @@ bool_to_scm (bool flag) {
   return scm_bool2scm (flag);
 }
 
-bool
+#ifndef GUILE_C
+int
 scm_to_bool (SCM flag) {
   return scm_scm2bool (flag);
 }
+#endif
 
 /******************************************************************************
 * Integers
 ******************************************************************************/
 
 #define SCM_ASSERT_INT(i,arg,rout) \
-  SCM_ASSERT (SCM_INUMP (i), i, arg, rout);
+  SCM_ASSERT (scm_is_int (i), i, arg, rout);
 
 SCM
 int_to_scm (int i) {
   return scm_long2scm ((long) i);
 }
 
+#ifndef GUILE_C
 int
 scm_to_int (SCM i) {
   return (int) scm_scm2long (i);
 }
+#endif
 
 /******************************************************************************
 * Floating point numbers
@@ -142,19 +166,19 @@ double_to_scm (double i) {
   return scm_double2scm (i);
 }
 
+#ifndef GUILE_C
 static double
 scm_to_double (SCM i) {
   return scm_scm2double (i);
 }
+#endif
 
 /******************************************************************************
 * Strings
 ******************************************************************************/
 
-#define SCM_MYSTRINGP(s) (SCM_NIMP (s) && SCM_STRINGP (s))
-
 #define SCM_ASSERT_STRING(s,arg,rout) \
-  SCM_ASSERT (SCM_MYSTRINGP (s), s, arg, rout)
+  SCM_ASSERT (scm_is_string (s), s, arg, rout)
 
 SCM
 string_to_scm (string s) {
@@ -184,7 +208,7 @@ scm_to_string (SCM s) {
 #define SCM_ASSERT_SYMBOL(s,arg,rout) \
   SCM_ASSERT (SCM_NFALSEP (scm_symbol_p (s)), s, arg, rout)
 
-static SCM
+SCM
 symbol_to_scm (string s) {
   char* _s= as_charp (s);
   SCM r= scm_symbol2scm (_s);
@@ -192,7 +216,7 @@ symbol_to_scm (string s) {
   return r;
 }
 
-static string
+string
 scm_to_symbol (SCM s) {
   guile_str_size_t len_r;
   char* _r= scm_scm2symbol (s, &len_r);
@@ -242,9 +266,7 @@ scm_is_tree (SCM t) {
 SCM
 tree_to_scm (tree t) {
   SCM tree_smob;
-  SCM_NEWCELL (tree_smob);
-  SCM_SETCDR (tree_smob, (SCM) ((void*) (new tree (t))));
-  SCM_SETCAR (tree_smob, tree_tag);
+  SET_SMOB (tree_smob, (void*) (new tree (t)), (SCM) tree_tag);
   return tree_smob;
 }
 
@@ -263,7 +285,7 @@ static scm_sizet
 free_tree (SCM tree_smob) {
   tree *ptr = (tree *) SCM_CDR (tree_smob);
   delete ptr;
-  return sizeof (tree); // should be replaced by total size of the tree
+  return 0;
 }
 
 static int
@@ -301,9 +323,10 @@ tree_ref (tree t, int i) {
   return t[i];
 }
 
-void
+tree
 tree_set (tree t, int i, tree u) {
   t[i]= u;
+  return u;
 }
 
 tree
@@ -316,6 +339,132 @@ tree_append (tree t1, tree t2) {
   return t1 * t2;
 }
 
+bool
+tree_active (tree t) {
+  path ip= obtain_ip (t);
+  return nil (ip) || last_item (ip) != DETACHED;
+}
+
+tree
+tree_child_insert (tree t, int pos, tree x) {
+  cout << "t= " << t << "\n";
+  cout << "x= " << x << "\n";
+  int i, n= N(t);
+  tree r (t, n+1);
+  for (i=0; i<pos; i++) r[i]= t[i];
+  r[pos]= x;
+  for (i=pos; i<n; i++) r[i+1]= t[i];
+  return r;
+}
+
+/******************************************************************************
+* Document modification routines
+******************************************************************************/
+
+static inline bool
+ip_attached (path ip) {
+  return nil (ip) || last_item (ip) != DETACHED;
+}
+
+tree
+tree_assign (tree r, tree t) {
+  path ip= obtain_ip (r);
+  //cout << "Assign " << r << ", " << t << " at " << ip << "\n";
+  if (ip_attached (ip)) {
+    path p= reverse (ip);
+    get_server()->get_editor()->assign (p, copy (t));
+    return get_server()->get_editor()->the_subtree (p);
+  }
+  else {
+    assign (r, t);
+    return r;
+  }
+}
+
+tree
+tree_insert (tree r, int pos, tree t) {
+  path ip= obtain_ip (r);
+  //cout << "Insert " << r << ", " << pos << ", " << t
+  //     << " at " << ip << "\n";
+  if (ip_attached (ip))
+    get_server()->get_editor()->insert (reverse (ip) * pos, copy (t));
+  else insert (r, pos, t);
+  return r;
+}
+
+tree
+tree_remove (tree r, int pos, int nr) {
+  path ip= obtain_ip (r);
+  //cout << "Remove " << r << ", " << pos << ", " << nr
+  //     << " at " << ip << "\n";
+  if (ip_attached (ip))
+    get_server()->get_editor()->remove (reverse (ip) * pos, nr);
+  else remove (r, pos, nr);
+  return r;
+}
+
+tree
+tree_split (tree r, int pos, int at) {
+  path ip= obtain_ip (r);
+  //cout << "Split " << r << ", " << pos << ", " << at
+  //     << " at " << ip << "\n";
+  if (ip_attached (ip))
+    get_server()->get_editor()->split (reverse (ip) * path (pos, at));
+  else split (r, pos, at);
+  return r;
+}
+
+tree
+tree_join (tree r, int pos) {
+  path ip= obtain_ip (r);
+  //cout << "Join " << r << ", " << pos << " at " << ip << "\n";
+  if (ip_attached (ip))
+    get_server()->get_editor()->join (reverse (ip) * pos);
+  else join (r, pos);
+  return r;
+}
+
+tree
+tree_assign_node (tree r, tree_label op) {
+  path ip= obtain_ip (r);
+  //cout << "Assign node " << r << ", " << tree (op) << " at " << ip << "\n";
+  if (ip_attached (ip))
+    get_server()->get_editor()->assign_node (reverse (ip), op);
+  else assign_node (r, op);
+  return r;
+}
+
+tree
+tree_insert_node (tree r, int pos, tree t) {
+  path ip= obtain_ip (r);
+  //cout << "Insert node " << r << ", " << pos << ", " << t
+  //     << " at " << ip << "\n";
+  if (ip_attached (ip)) {
+    path p= reverse (ip);
+    get_server()->get_editor()->insert_node (p * pos, copy (t));
+    return get_server()->get_editor()->the_subtree (p);
+  }
+  else {
+    insert_node (r, pos, t);
+    return r;
+  }
+}
+
+tree
+tree_remove_node (tree r, int pos) {
+  path ip= obtain_ip (r);
+  //cout << "Remove node " << r << ", " << pos << " at " << ip << "\n";
+  if (ip_attached (ip)) {
+    path p= reverse (ip);
+    get_server()->get_editor()->remove_node (p * pos);
+    return get_server()->get_editor()->the_subtree (p);
+  }
+  else {
+    remove_node (r, pos);
+    return r;
+  }
+}
+
 /******************************************************************************
 * Scheme trees
 ******************************************************************************/
@@ -329,8 +478,10 @@ scheme_tree_to_scm (scheme_tree t) {
     if (s == "#t") return SCM_BOOL_T;
     if (s == "#f") return SCM_BOOL_F;
     if (is_int (s)) return int_to_scm (as_int (s));
-    if ((N(s)>=2) && (s[0]=='\42') && (s[N(s)-1]=='\42'))
-      return string_to_scm (s (1, N(s)-1));
+    if (is_quoted (s))
+      return string_to_scm (scm_unquote (s));
+    //if ((N(s)>=2) && (s[0]=='\42') && (s[N(s)-1]=='\42'))
+    //return string_to_scm (s (1, N(s)-1));
     return symbol_to_scm (s);
   }
   else {
@@ -352,9 +503,10 @@ scm_to_scheme_tree (SCM p) {
     }
     return t;
   }
-  if (gh_symbol_p (p)) return scm_to_symbol (p);
-  if (scm_is_string (p)) return "\"" * scm_to_string (p) * "\"";
-  if (SCM_INUMP (p)) return as_string (scm_to_int (p));
+  if (scm_is_symbol (p)) return scm_to_symbol (p);
+  if (scm_is_string (p)) return scm_quote (scm_to_string (p));
+  //if (scm_is_string (p)) return "\"" * scm_to_string (p) * "\"";
+  if (scm_is_int (p)) return as_string ((int) scm_to_int (p));
   if (scm_is_bool (p)) return (scm_to_bool (p)? string ("#t"): string ("#f"));
   if (scm_is_tree (p)) return tree_to_scheme_tree (scm_to_tree (p));
   return "?";
@@ -364,15 +516,28 @@ scm_to_scheme_tree (SCM p) {
 * Content
 ******************************************************************************/
 
+bool
+scm_is_content (SCM p) {
+  if (scm_is_string (p) || scm_is_tree (p)) return true;
+  else if (!scm_is_pair (p) || !scm_is_symbol (SCM_CAR (p))) return false;
+  else {
+    for (p= SCM_CDR (p); !scm_is_null (p); p= SCM_CDR (p))
+      if (!scm_is_content (SCM_CAR (p))) return false;
+    return true;
+  }
+}
+
 #define content tree
-#define SCM_ASSERT_CONTENT(p,arg,rout)
+#define SCM_ASSERT_CONTENT(p,arg,rout) \
+  SCM_ASSERT (scm_is_content (p), p, arg, rout)
 #define content_to_scm tree_to_scm
 
 tree
 scm_to_content (SCM p) {
+  if (scm_is_string (p)) return scm_to_string (p);
   if (scm_is_tree (p)) return scm_to_tree (p);
-  if (scm_is_list (p)) {
-    if (scm_is_null (p) || (!gh_symbol_p (SCM_CAR (p)))) return "?";
+  if (scm_is_pair (p)) {
+    if (!scm_is_symbol (SCM_CAR (p))) return "?";
     tree t (make_tree_label (scm_to_symbol (SCM_CAR (p))));
     p= SCM_CDR (p);
     while (!scm_is_null (p)) {
@@ -381,13 +546,15 @@ scm_to_content (SCM p) {
     }
     return t;
   }
-  if (gh_symbol_p (p)) return scm_to_symbol (p);
-  if (scm_is_string (p)) return scm_to_string (p);
-  if (SCM_INUMP (p)) return as_string (scm_to_int (p));
-  if (scm_is_bool (p)) return (scm_to_bool (p)? string ("#t"): string ("#f"));
   return "?";
 }
 
+static SCM
+contentP (SCM t) {
+  bool b= scm_is_content (t);
+  return bool_to_scm (b);
+}
+
 /******************************************************************************
 * Paths
 ******************************************************************************/
@@ -395,7 +562,7 @@ scm_to_content (SCM p) {
 bool
 scm_is_path (SCM p) {
   if (scm_is_null (p)) return true;
-  else return SCM_INUMP (SCM_CAR (p)) && scm_is_path (SCM_CDR (p));
+  else return scm_is_int (SCM_CAR (p)) && scm_is_path (SCM_CDR (p));
 }
 
 #define SCM_ASSERT_PATH(p,arg,rout) \
@@ -410,58 +577,62 @@ path_to_scm (path p) {
 path
 scm_to_path (SCM p) {
   if (scm_is_null (p)) return path ();
-  else return path (scm_to_int (SCM_CAR (p)), scm_to_path (SCM_CDR (p)));
+  else return path ((int) scm_to_int (SCM_CAR (p)), scm_to_path (SCM_CDR (p)));
 }
 
 /******************************************************************************
-* Displays
+* Observers
 ******************************************************************************/
 
-static long display_tag;
+static long observer_tag;
 
-#define scm_is_display(t) \
-  (SCM_NIMP (t) && (((long) SCM_CAR (t)) == display_tag))
-#define SCM_ASSERT_DISPLAY(t,arg,rout) \
-  SCM_ASSERT (scm_is_display (t), t, arg, rout)
+#define scm_is_observer(o) \
+  (SCM_NIMP (o) && (((long) SCM_CAR (o)) == observer_tag))
+#define SCM_ASSERT_OBSERVER(o,arg,rout) \
+  SCM_ASSERT (scm_is_observer (o), o, arg, rout)
 
 /*static*/ SCM
-display_to_scm (display t) {
-  SCM display_smob;
-  SCM_NEWCELL (display_smob);
-  SCM_SETCDR (display_smob, (SCM) ((void*) (new display (t))));
-  SCM_SETCAR (display_smob, display_tag);
-  return display_smob;
+observer_to_scm (observer o) {
+  SCM observer_smob;
+  SET_SMOB (observer_smob, (void*) (new observer (o)), (SCM) observer_tag);
+  return observer_smob;
 }
 
-static display
-scm_to_display (SCM display_smob) {
-  return *((display*) SCM_CDR (display_smob));
+static observer
+scm_to_observer (SCM observer_smob) {
+  return *((observer*) SCM_CDR (observer_smob));
 }
 
 static SCM
-mark_display (SCM display_smob) {
-  (void) display_smob;
+mark_observer (SCM observer_smob) {
+  (void) observer_smob;
   return SCM_BOOL_F;
 }
 
 static scm_sizet
-free_display (SCM display_smob) {
-  display *ptr = (display *) SCM_CDR (display_smob);
+free_observer (SCM observer_smob) {
+  observer *ptr = (observer *) SCM_CDR (observer_smob);
   delete ptr;
-  return sizeof (display); // should be replaced by total size of the display
+  return 0;
 }
 
 static int
-print_display (SCM display_smob, SCM port, scm_print_state *pstate) {
-  (void) display_smob; (void) pstate;
-  string s= "<display>";
+print_observer (SCM observer_smob, SCM port, scm_print_state *pstate) {
+  (void) observer_smob; (void) pstate;
+  string s= "<observer>";
   scm_display (string_to_scm (s), port);
   return 1;
 }
 
 static SCM
-cmp_display (SCM t1, SCM t2) {
-  return scm_bool2scm (scm_to_display (t1) == scm_to_display (t2));
+cmp_observer (SCM o1, SCM o2) {
+  return scm_bool2scm (scm_to_observer (o1) == scm_to_observer (o2));
+}
+
+static SCM
+observerP (SCM t) {
+  bool b= scm_is_observer (t);
+  return bool_to_scm (b);
 }
 
 /******************************************************************************
@@ -470,22 +641,20 @@ cmp_display (SCM t1, SCM t2) {
 
 static long widget_tag;
 
-#define SCM_WIDGETP(t) \
-  (SCM_NIMP (t) && (((long) SCM_CAR (t)) == widget_tag))
-#define SCM_ASSERT_WIDGET(t,arg,rout) \
-  SCM_ASSERT (scm_is_widget (t), t, arg, rout)
+#define SCM_WIDGETP(wid) \
+  (SCM_NIMP (wid) && (((long) SCM_CAR (wid)) == widget_tag))
+#define SCM_ASSERT_WIDGET(wid,arg,rout) \
+  SCM_ASSERT (scm_is_widget (wid), wid, arg, rout)
 
 bool
-scm_is_widget (SCM t) {
-  return SCM_WIDGETP (t);
+scm_is_widget (SCM wid) {
+  return SCM_WIDGETP (wid);
 }
 
 static SCM
-widget_to_scm (widget t) {
+widget_to_scm (widget wid) {
   SCM widget_smob;
-  SCM_NEWCELL (widget_smob);
-  SCM_SETCDR (widget_smob, (SCM) ((void*) (new widget (t))));
-  SCM_SETCAR (widget_smob, widget_tag);
+  SET_SMOB (widget_smob, (void*) (new widget (wid)), (SCM) widget_tag);
   return widget_smob;
 }
 
@@ -504,7 +673,7 @@ static scm_sizet
 free_widget (SCM widget_smob) {
   widget *ptr = (widget *) SCM_CDR (widget_smob);
   delete ptr;
-  return sizeof (widget); // should be replaced by total size of the widget
+  return 0;
 }
 
 static int
@@ -516,61 +685,66 @@ print_widget (SCM widget_smob, SCM port, scm_print_state *pstate) {
 }
 
 static SCM
-cmp_widget (SCM t1, SCM t2) {
-  return scm_bool2scm (scm_to_widget (t1) == scm_to_widget (t2));
+cmp_widget (SCM wid1, SCM wid2) {
+  return scm_bool2scm (scm_to_widget (wid1) == scm_to_widget (wid2));
 }
 
 /******************************************************************************
 * Widget factory
 ******************************************************************************/
 
-static long make_widget_tag;
+typedef promise<widget> promise_widget;
 
-#define SCM_ASSERT_MAKE_WIDGET(t,arg,rout) \
-  SCM_ASSERT (scm_is_make_widget (t), t, arg, rout)
+static long promise_widget_tag;
 
-#define scm_is_make_widget(t) \
-  (SCM_NIMP (t) && (((long) SCM_CAR (t)) == make_widget_tag))
+#define SCM_ASSERT_PROMISE_WIDGET(pw,arg,rout) \
+  SCM_ASSERT (scm_is_promise_widget (pw), pw, arg, rout)
+
+#define scm_is_promise_widget(pw) \
+  (SCM_NIMP (pw) && (((long) SCM_CAR (pw)) == promise_widget_tag))
 
 static SCM
-make_widget_to_scm (make_widget t) {
-  SCM make_widget_smob;
-  SCM_NEWCELL (make_widget_smob);
-  SCM_SETCDR (make_widget_smob, (SCM) ((void*) (new make_widget (t))));
-  SCM_SETCAR (make_widget_smob, make_widget_tag);
-  return make_widget_smob;
+promise_widget_to_scm (promise_widget pw) {
+  SCM promise_widget_smob;
+  SET_SMOB (promise_widget_smob,
+	    (void*) (new promise_widget (pw)),
+	    (SCM) promise_widget_tag);
+  return promise_widget_smob;
 }
 
-static make_widget
-scm_to_make_widget (SCM make_widget_smob) {
-  return *((make_widget*) SCM_CDR (make_widget_smob));
+static promise_widget
+scm_to_promise_widget (SCM promise_widget_smob) {
+  return *((promise_widget*) SCM_CDR (promise_widget_smob));
 }
 
 static SCM
-mark_make_widget (SCM make_widget_smob) {
-  (void) make_widget_smob;
+mark_promise_widget (SCM promise_widget_smob) {
+  (void) promise_widget_smob;
   return SCM_BOOL_F;
 }
 
 static scm_sizet
-free_make_widget (SCM make_widget_smob) {
-  make_widget *ptr = (make_widget *) SCM_CDR (make_widget_smob);
+free_promise_widget (SCM promise_widget_smob) {
+  promise_widget *ptr = (promise_widget *) SCM_CDR (promise_widget_smob);
   delete ptr;
   // should be replaced by total size of the widget factory
-  return sizeof (make_widget);
+  return 0;
 }
 
 static int
-print_make_widget (SCM make_widget_smob, SCM port, scm_print_state *pstate) {
-  (void) make_widget_smob; (void) pstate;
-  string s= "<make-widget>";
+print_promise_widget (SCM promise_widget_smob, SCM port,
+		      scm_print_state *pstate)
+{
+  (void) promise_widget_smob; (void) pstate;
+  string s= "<promise-widget>";
   scm_display (string_to_scm (s), port);
   return 1;
 }
 
 static SCM
-cmp_make_widget (SCM t1, SCM t2) {
-  return scm_bool2scm (scm_to_make_widget (t1) == scm_to_make_widget (t2));
+cmp_promise_widget (SCM pw1, SCM pw2) {
+  return scm_bool2scm (scm_to_promise_widget (pw1) ==
+		       scm_to_promise_widget (pw2));
 }
 
 /******************************************************************************
@@ -579,17 +753,15 @@ cmp_make_widget (SCM t1, SCM t2) {
 
 static long command_tag;
 
-#define scm_is_command(t) \
-  (SCM_NIMP (t) && (((long) SCM_CAR (t)) == command_tag))
-#define SCM_ASSERT_COMMAND(t,arg,rout) \
-  SCM_ASSERT (scm_is_command (t), t, arg, rout)
+#define scm_is_command(cmd) \
+  (SCM_NIMP (cmd) && (((long) SCM_CAR (cmd)) == command_tag))
+#define SCM_ASSERT_COMMAND(cmd,arg,rout) \
+  SCM_ASSERT (scm_is_command (cmd), cmd, arg, rout)
 
 static SCM
-command_to_scm (command t) {
+command_to_scm (command cmd) {
   SCM command_smob;
-  SCM_NEWCELL (command_smob);
-  SCM_SETCDR (command_smob, (SCM) ((void*) (new command (t))));
-  SCM_SETCAR (command_smob, command_tag);
+  SET_SMOB (command_smob, (void*) (new command (cmd)), (SCM) command_tag);
   return command_smob;
 }
 
@@ -608,7 +780,7 @@ static scm_sizet
 free_command (SCM command_smob) {
   command *ptr = (command *) SCM_CDR (command_smob);
   delete ptr;
-  return sizeof (command); // should be replaced by total size of the command
+  return 0;
 }
 
 static int
@@ -620,8 +792,8 @@ print_command (SCM command_smob, SCM port, scm_print_state *pstate) {
 }
 
 static SCM
-cmp_command (SCM t1, SCM t2) {
-  return scm_bool2scm (scm_to_command (t1) == scm_to_command (t2));
+cmp_command (SCM cmd1, SCM cmd2) {
+  return scm_bool2scm (scm_to_command (cmd1) == scm_to_command (cmd2));
 }
 
 /******************************************************************************
@@ -633,7 +805,7 @@ static long url_tag;
 #define SCM_URLP(u) \
   (SCM_NIMP (u) && (((long) SCM_CAR (u)) == url_tag))
 #define SCM_ASSERT_URL(u,arg,rout) \
-  SCM_ASSERT (scm_is_url (u) || SCM_MYSTRINGP (u), u, arg, rout)
+  SCM_ASSERT (scm_is_url (u) || scm_is_string (u), u, arg, rout)
 
 bool
 scm_is_url (SCM u) {
@@ -643,9 +815,7 @@ scm_is_url (SCM u) {
 SCM
 url_to_scm (url u) {
   SCM url_smob;
-  SCM_NEWCELL (url_smob);
-  SCM_SETCDR (url_smob, (SCM) ((void*) (new url (u))));
-  SCM_SETCAR (url_smob, url_tag);
+  SET_SMOB (url_smob, (void*) (new url (u)), (SCM) url_tag);
   return url_smob;
 }
 
@@ -666,7 +836,7 @@ static scm_sizet
 free_url (SCM url_smob) {
   url *ptr = (url *) SCM_CDR (url_smob);
   delete ptr;
-  return sizeof (url); // should be replaced by total size of the url
+  return 0;
 }
 
 static int
@@ -683,24 +853,114 @@ cmp_url (SCM u1, SCM u2) {
   return scm_bool2scm (scm_to_url (u1) == scm_to_url (u2));
 }
 
+static SCM
+urlP (SCM t) {
+  bool b= scm_is_url (t);
+  return bool_to_scm (b);
+}
+
 url url_concat (url u1, url u2) { return u1 * u2; }
 url url_or (url u1, url u2) { return u1 | u2; }
 void string_save (string s, url u) { (void) save_string (u, s); }
-string string_load (url u) { string s; (void) load_string (u, s); return s; }
+string string_load (url u) {
+  string s; (void) load_string (u, s, false); return s; }
 url url_ref (url u, int i) { return u[i]; }
 
+/******************************************************************************
+* Table types
+******************************************************************************/
+
+typedef hashmap<string,string> table_string_string;
+
+static bool
+scm_is_table_string_string (SCM p) {
+  if (scm_is_null (p)) return true;
+  else if (!scm_is_pair (p)) return false;
+  else {
+    SCM f= SCM_CAR (p);
+    return scm_is_pair (f) &&
+           scm_is_string (SCM_CAR (f)) &&
+           scm_is_string (SCM_CDR (f)) &&
+           scm_is_table_string_string (SCM_CDR (p));
+  }
+}
+
+#define SCM_ASSERT_TABLE_STRING_STRING(p,arg,rout) \
+  SCM_ASSERT (scm_is_table_string_string (p), p, arg, rout)
+
+SCM
+table_string_string_to_scm (hashmap<string,string> t) {
+  SCM p= SCM_NULL;
+  iterator<string> it= iterate (t);
+  while (it->busy ()) {
+    string s= it->next ();
+    SCM n= scm_cons (string_to_scm (s), string_to_scm (t[s]));
+    p= scm_cons (n, p);
+  }
+  return p;
+}
+
+hashmap<string,string>
+scm_to_table_string_string (SCM p) {
+  hashmap<string,string> t;
+  while (!scm_is_null (p)) {
+    SCM n= SCM_CAR (p);
+    t (scm_to_string (SCM_CAR (n)))= scm_to_string (SCM_CDR (n));
+    p= SCM_CDR (p);
+  }
+  return t;
+}
+
+#define scm_is_solution scm_is_table_string_string
+#define SCM_ASSERT_SOLUTION(p,arg,rout) \
+  SCM_ASSERT (scm_is_solution(p), p, arg, rout)
+#define solution_to_scm table_string_string_to_scm
+#define scm_to_solution scm_to_table_string_string
+
 /******************************************************************************
 * Several array types
 ******************************************************************************/
 
+typedef array<int> array_int;
 typedef array<string> array_string;
 typedef array<tree> array_tree;
 typedef array<widget> array_widget;
 
+static bool
+scm_is_array_int (SCM p) {
+  if (scm_is_null (p)) return true;
+  else return scm_is_pair (p) &&
+	      scm_is_int (SCM_CAR (p)) &&
+	      scm_is_array_int (SCM_CDR (p));
+}
+
+#define SCM_ASSERT_ARRAY_INT(p,arg,rout) \
+  SCM_ASSERT (scm_is_array_int (p), p, arg, rout)
+
+/* static */ SCM
+array_int_to_scm (array<int> a) {
+  int i, n= N(a);
+  SCM p= SCM_NULL;
+  for (i=n-1; i>=0; i--) p= scm_cons (int_to_scm (a[i]), p);
+  return p;
+}
+
+/* static */ array<int>
+scm_to_array_int (SCM p) {
+  array<int> a;
+  while (!scm_is_null (p)) {
+    a << ((int) scm_to_int (SCM_CAR (p)));
+    p= SCM_CDR (p);
+  }
+  return a;
+}
+
 static bool
 scm_is_array_string (SCM p) {
   if (scm_is_null (p)) return true;
-  else return SCM_MYSTRINGP (SCM_CAR (p)) && scm_is_array_string (SCM_CDR (p));
+  else return scm_is_pair (p) && 
+	      scm_is_string (SCM_CAR (p)) &&
+	      scm_is_array_string (SCM_CDR (p));
 }
 
 #define SCM_ASSERT_ARRAY_STRING(p,arg,rout) \
@@ -724,10 +984,17 @@ scm_to_array_string (SCM p) {
   return a;
 }
 
+#define scm_is_property scm_is_array_string
+#define SCM_ASSERT_PROPERTY(p,arg,rout) SCM_ASSERT_ARRAY_STRING (p,arg,rout)
+#define property_to_scm array_string_to_scm
+#define scm_to_property scm_to_array_string
+
 static bool
 scm_is_array_tree (SCM p) {
   if (scm_is_null (p)) return true;
-  else return SCM_TREEP (SCM_CAR (p)) && scm_is_array_tree (SCM_CDR (p));
+  else return scm_is_pair (p) && 
+	      SCM_TREEP (SCM_CAR (p)) &&
+	      scm_is_array_tree (SCM_CDR (p));
 }
 
 #define SCM_ASSERT_ARRAY_TREE(p,arg,rout) \
@@ -754,7 +1021,9 @@ scm_to_array_tree (SCM p) {
 static bool
 scm_is_array_widget (SCM p) {
   if (scm_is_null (p)) return true;
-  else return scm_is_widget (SCM_CAR (p)) && scm_is_array_widget (SCM_CDR (p));
+  else return scm_is_pair (p) &&
+	      scm_is_widget (SCM_CAR (p)) &&
+	      scm_is_array_widget (SCM_CDR (p));
 }
 
 #define SCM_ASSERT_ARRAY_WIDGET(p,arg,rout) \
@@ -778,6 +1047,139 @@ scm_to_array_widget (SCM p) {
   return a;
 }
 
+static bool
+scm_is_properties (SCM p) {
+  if (scm_is_null (p)) return true;
+  else return scm_is_pair (p) &&
+	      scm_is_property (SCM_CAR (p)) &&
+	      scm_is_properties (SCM_CDR (p));
+}
+
+#define SCM_ASSERT_PROPERTIES(p,arg,rout) \
+  SCM_ASSERT (scm_is_properties (p), p, arg, rout)
+
+SCM
+properties_to_scm (array<property> a) {
+  int i, n= N(a);
+  SCM p= SCM_NULL;
+  for (i=n-1; i>=0; i--) p= scm_cons (property_to_scm (a[i]), p);
+  return p;
+}
+
+array<property>
+scm_to_properties (SCM p) {
+  array<property> a;
+  while (!scm_is_null (p)) {
+    a << scm_to_property (SCM_CAR (p));
+    p= SCM_CDR (p);
+  }
+  return a;
+}
+
+static bool
+scm_is_solutions (SCM p) {
+  if (scm_is_null (p)) return true;
+  else return scm_is_pair (p) &&
+	      scm_is_solution (SCM_CAR (p)) &&
+	      scm_is_solutions (SCM_CDR (p));
+}
+
+#define SCM_ASSERT_SOLUTIONS(p,arg,rout) \
+  SCM_ASSERT (scm_is_solutions (p), p, arg, rout)
+
+SCM
+solutions_to_scm (array<solution> a) {
+  int i, n= N(a);
+  SCM p= SCM_NULL;
+  for (i=n-1; i>=0; i--) p= scm_cons (solution_to_scm (a[i]), p);
+  return p;
+}
+
+array<solution>
+scm_to_solutions (SCM p) {
+  array<solution> a;
+  while (!scm_is_null (p)) {
+    a << scm_to_solution (SCM_CAR (p));
+    p= SCM_CDR (p);
+  }
+  return a;
+}
+
+/******************************************************************************
+* List types
+******************************************************************************/
+
+typedef list<string> list_string;
+
+bool
+scm_is_list_string (SCM p) {
+  if (scm_is_null (p)) return true;
+  else return scm_is_pair (p) &&
+	      scm_is_string (SCM_CAR (p)) &&
+	      scm_is_list_string (SCM_CDR (p));
+}
+
+#define SCM_ASSERT_LIST_STRING(p,arg,rout) \
+  SCM_ASSERT (scm_is_list_string (p), p, arg, rout)
+
+SCM
+list_string_to_scm (list_string l) {
+  if (nil (l)) return SCM_NULL;
+  return scm_cons (string_to_scm (l->item),
+		   list_string_to_scm (l->next));
+}
+
+list_string
+scm_to_list_string (SCM p) {
+  if (scm_is_null (p)) return list_string ();
+  return list_string (scm_to_string (SCM_CAR (p)),
+		      scm_to_list_string (SCM_CDR (p)));
+}
+
+typedef list<tree> list_tree;
+
+bool
+scm_is_list_tree (SCM p) {
+  if (scm_is_null (p)) return true;
+  else return scm_is_pair (p) &&
+	      scm_is_tree (SCM_CAR (p)) &&
+	      scm_is_list_tree (SCM_CDR (p));
+}
+
+#define SCM_ASSERT_LIST_TREE(p,arg,rout) \
+  SCM_ASSERT (scm_is_list_tree (p), p, arg, rout)
+
+SCM
+list_tree_to_scm (list_tree l) {
+  if (nil (l)) return SCM_NULL;
+  return scm_cons (tree_to_scm (l->item),
+		   list_tree_to_scm (l->next));
+}
+
+list_tree
+scm_to_list_tree (SCM p) {
+  if (scm_is_null (p)) return list_tree ();
+  return list_tree (scm_to_tree (SCM_CAR (p)),
+		    scm_to_list_tree (SCM_CDR (p)));
+}
+
+/******************************************************************************
+* Other wrapper types
+******************************************************************************/
+
+#define SCM_ASSERT_COLLECTION(p,arg,rout) \
+  SCM_ASSERT (scm_is_array_string (p), p, arg, rout)
+
+SCM
+collection_to_scm (collection ss) {
+  return array_string_to_scm (as_strings (ss));
+}
+
+collection
+scm_to_collection (SCM p) {
+  return as_collection (scm_to_array_string (p));
+}
+
 /******************************************************************************
 * Initialization
 ******************************************************************************/
@@ -791,21 +1193,21 @@ initialize_glue () {
   scm_set_smob_free (tree_tag, free_tree);
   scm_set_smob_print (tree_tag, print_tree);
   scm_set_smob_equalp (tree_tag, cmp_tree);
-  display_tag= scm_make_smob_type ("display", 0);
-  scm_set_smob_mark (display_tag, mark_display);
-  scm_set_smob_free (display_tag, free_display);
-  scm_set_smob_print (display_tag, print_display);
-  scm_set_smob_equalp (display_tag, cmp_display);
+  observer_tag= scm_make_smob_type ("observer", 0);
+  scm_set_smob_mark (observer_tag, mark_observer);
+  scm_set_smob_free (observer_tag, free_observer);
+  scm_set_smob_print (observer_tag, print_observer);
+  scm_set_smob_equalp (observer_tag, cmp_observer);
   widget_tag= scm_make_smob_type ("widget", 0);
   scm_set_smob_mark (widget_tag, mark_widget);
   scm_set_smob_free (widget_tag, free_widget);
   scm_set_smob_print (widget_tag, print_widget);
   scm_set_smob_equalp (widget_tag, cmp_widget);
-  make_widget_tag= scm_make_smob_type ("make-widget", 0);
-  scm_set_smob_mark (make_widget_tag, mark_make_widget);
-  scm_set_smob_free (make_widget_tag, free_make_widget);
-  scm_set_smob_print (make_widget_tag, print_make_widget);
-  scm_set_smob_equalp (make_widget_tag, cmp_make_widget);
+  promise_widget_tag= scm_make_smob_type ("promise-widget", 0);
+  scm_set_smob_mark (promise_widget_tag, mark_promise_widget);
+  scm_set_smob_free (promise_widget_tag, free_promise_widget);
+  scm_set_smob_print (promise_widget_tag, print_promise_widget);
+  scm_set_smob_equalp (promise_widget_tag, cmp_promise_widget);
   command_tag= scm_make_smob_type ("command", 0);
   scm_set_smob_mark (command_tag, mark_command);
   scm_set_smob_free (command_tag, free_command);
@@ -816,7 +1218,10 @@ initialize_glue () {
   scm_set_smob_free (url_tag, free_url);
   scm_set_smob_print (url_tag, print_url);
   scm_set_smob_equalp (url_tag, cmp_url);
-  gh_new_procedure ("tree?", (FN) treeP, 1, 0, 0);
+  scm_new_procedure ("tree?", (FN) treeP, 1, 0, 0);
+  scm_new_procedure ("tm?", (FN) contentP, 1, 0, 0);
+  scm_new_procedure ("observer?", (FN) observerP, 1, 0, 0);
+  scm_new_procedure ("url?", (FN) urlP, 1, 0, 0);
   initialize_glue_basic ();
   initialize_glue_editor ();
   initialize_glue_server ();
@@ -828,16 +1233,17 @@ scm_smobfuns tree_smob_funcs = {
   mark_tree, free_tree, print_tree, cmp_tree
 };
 
-scm_smobfuns display_smob_funcs = {
-  mark_display, free_display, print_display, cmp_display
+scm_smobfuns observer_smob_funcs = {
+  mark_observer, free_observer, print_observer, cmp_observer
 };
 
 scm_smobfuns widget_smob_funcs = {
   mark_widget, free_widget, print_widget, cmp_widget
 };
 
-scm_smobfuns make_widget_smob_funcs = {
-  mark_make_widget, free_make_widget, print_make_widget, cmp_make_widget
+scm_smobfuns promise_widget_smob_funcs = {
+  mark_promise_widget, free_promise_widget,
+  print_promise_widget, cmp_promise_widget
 };
 
 scm_smobfuns command_smob_funcs = {
@@ -851,12 +1257,15 @@ scm_smobfuns url_smob_funcs = {
 void
 initialize_glue () {
   tree_tag= scm_newsmob (&tree_smob_funcs);
-  display_tag= scm_newsmob (&display_smob_funcs);
+  observer_tag= scm_newsmob (&observer_smob_funcs);
   widget_tag= scm_newsmob (&widget_smob_funcs);
-  make_widget_tag= scm_newsmob (&make_widget_smob_funcs);
+  promise_widget_tag= scm_newsmob (&promise_widget_smob_funcs);
   command_tag= scm_newsmob (&command_smob_funcs);
   url_tag= scm_newsmob (&url_smob_funcs);
-  gh_new_procedure ("tree?", (FN) treeP, 1, 0, 0);
+  scm_new_procedure ("tree?", (FN) treeP, 1, 0, 0);
+  scm_new_procedure ("tm?", (FN) contentP, 1, 0, 0);
+  scm_new_procedure ("observer?", (FN) observerP, 1, 0, 0);
+  scm_new_procedure ("url?", (FN) urlP, 1, 0, 0);
   initialize_glue_basic ();
   initialize_glue_editor ();
   initialize_glue_server ();
